import os
import json
from cyclonefw.core.utils import getLogger

logger = getLogger("config")


class Config(dict):
    """
    配置工具类, 使用方式:

    from cyclonefw.utils import Config

    conf = Config("conf.json")

    conf = Config({
        "data": 123
    })
    
    conf = Config('''{
        "data": 123
    }''')

    print(conf["data"])

    conf.reload()

    print(conf["data"])

    """

    def __init__(self, filenameOrContent):
        """
        创建配置对象
        :param filenameOrContent:
        """
        super(Config, self).__init__(self)
        logger.info("创建配置对象: {}".format(filenameOrContent))
        self.filenameOrContent = filenameOrContent
        self.data = {}
        self.reload()
        pass

    def reload(self):
        """
        配置重载
        :return:
        """
        try:
            if isinstance(self.filenameOrContent, dict):
                self.data = self.filenameOrContent

            if isinstance(self.filenameOrContent, str):
                self.filenameOrContent = self.filenameOrContent.strip()

                if self.filenameOrContent.startswith("{"):
                    self.data = json.loads(self.filenameOrContent)
                    return

                if not os.path.exists(self.filenameOrContent):
                    logger.error("配置文件不存在: {}".format(self.filenameOrContent))
                    self.data = {}
                    return

                with open(self.filenameOrContent, "r") as f:
                    fileContent = f.read()
                    self.data = json.loads(fileContent)
                    return

        except Exception as e:
            logger.error("配置文件载入失败: {}".format(self.filenameOrContent))
            logger.error(e)
        else:
            logger.info("配置载入成功: {}".format(self.data))
        pass

    def __getitem__(self, key):
        """
        获取配置内容
        :param key:
        :return:
        """
        if key in self.data:
            return self.data[key]
        return None



from flask import abort
import cyclonefw
import os, logging
from logging.handlers import RotatingFileHandler
import requests


def redirect_errors_to_flask(func):
    """
    This decorator function will capture all Pythonic errors and return them as flask errors.

    If you are looking to disable this functionality, please remove this decorator from the `apply_transforms()` module
    under the ImageProcessor class.
    """

    def inner(*args, **kwargs):
        try:
            # run the function
            return func(*args, **kwargs)
        except ValueError as ve:
            if 'pic should be 2 or 3 dimensional' in str(ve):
                abort(400, "Invalid input, please ensure the input is either "
                           "a grayscale or a colour image.")
        except TypeError as te:
            if 'bytes or ndarray' in str(te):
                abort(400, "Invalid input format, please make sure the input file format "
                           " is a common image format such as JPG or PNG.")

    return inner


class AppFilter(logging.Filter):
    def __init__(self):
        pass

    def filter(self, record):
        record.app_name = cyclonefw.core.app.context.app_name
        return True


def getLogger(name="cyclonefw", formatter="[%(asctime)s] [%(pathname)s(line:%(lineno)d)] [%(levelname)s] %(message)s",
              logPath=None, level=logging.DEBUG):
    """
    获取logger

    日志记录使用如下格式: %s 等作为占位符

    print "My name is %s and weight is %d kg!" % ('Zara', 21)

    :param name: 日志名称
    :param formatter: 日志格式
    :param logPath: 日志目录, 默认为 logs目录下, 使用 . 可以指定当前执行目录
    :param level: 日志级别
    :return:
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    if not logger.hasHandlers():
        consolelog = logging.StreamHandler()
        consolelog.setFormatter(logging.Formatter(formatter))
        logger.addHandler(consolelog)
        logger.addFilter(AppFilter())

    if logPath != None:
        logger.addHandler(logHandler(name, formatter, logPath))

    return logger


def logHandler(name, logPath="logs",
               formatter="[%(asctime)s] [%(pathname)s(line:%(lineno)d)] [%(levelname)s] %(message)s"):
    if not os.path.exists(logPath):
        os.makedirs(logPath, exist_ok=True)
    handler = CyclonefwRotatingFileHandler("{}/{}.log".format(logPath, name), "a", 1024 * 1024 * 20, 10)
    handler.setFormatter(logging.Formatter(formatter))
    return handler


class CyclonefwRotatingFileHandler(RotatingFileHandler):
    def emit(self, record):
        if hasattr(record, "trace_id"):
            self.setFormatter(logging.Formatter("[%(asctime)s] [%(trace_id)s] [%(app_name)s] [%(pathname)s(line:%(lineno)d)] [%(levelname)s] %(message)s"))

        super(CyclonefwRotatingFileHandler, self).emit(record)


def getHostnameAndIp():
    """
    获取hostname及ip地址
    :return: hostname, ip
    """
    import socket
    hostname = socket.gethostname()
    ip = socket.gethostbyname(hostname)
    return hostname, ip


class WrapperDict(dict):
    def __init__(self, seq=None, **kwargs):
        self.logger = getLogger()
        super(WrapperDict, self).__init__(seq, **kwargs)

    def __getitem__(self, item):
        value = dict.__getitem__(self, item)
        if item == "b64":
            if isinstance(value, str) and value.startswith("http"):
                try:
                    dict.__setitem__(self, item, requests.get(value, timeout=600).text)
                except Exception as e:
                    self.logger.error("get b64 from url failed: %s" % value)
                    self.logger.error(e)

        if item == "b64s":
            if isinstance(value, list) and len(value) > 0:
                if isinstance(value[0], str) and value[0].startswith("http"):
                    for i in range(len(value)):
                        try:
                            value[i] = requests.get(value[i], timeout=600).text
                        except Exception as e:
                            self.logger.error("get b64 from url failed: %s" % value)
                            self.logger.error(e)

        return dict.__getitem__(self, item)



#
# Copyright 2018-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import time
from flask import Flask, request, Response
from flask_restx import Api, Namespace
from flask_cors import CORS
from .default_config import API_TITLE, API_DESC, API_VERSION
from .utils import getLogger, logHandler, WrapperDict
import logging
import json
import requests
from werkzeug.local import LocalStack

from opentelemetry import trace
from opentelemetry.instrumentation.flask import FlaskInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.exporter.jaeger.proto import grpc
from opentelemetry.sdk.trace.export import (BatchSpanProcessor, ConsoleSpanExporter)
from opentelemetry.sdk.resources import Resource
from opentelemetry.exporter.zipkin.proto.http import ZipkinExporter

exporter_type = os.getenv('EXPORTER_TYPE', 'console')
exporter_url = os.getenv('EXPORTER_URL', '')
service_name = os.getenv('SERVICE_NAME', 'cyclone-algorithm-component')
# 定义trace
exporter = ConsoleSpanExporter()
if exporter_type.lower() == 'jaeger':
    exporter = grpc.JaegerExporter(
        collector_endpoint=exporter_url,
        insecure=True,
    )
elif exporter_type.lower() == 'zipkin':
    exporter = ZipkinExporter(endpoint=exporter_url)

resource = Resource(attributes={"service.name": service_name})

trace.set_tracer_provider(TracerProvider(resource=resource))
trace.get_tracer_provider().add_span_processor(BatchSpanProcessor(exporter))

# 重新定义flask日志
logging.getLogger("werkzeug").addHandler(logHandler("flask"))
# 定义默认logger
getLogger().addHandler(logHandler("cyclonefw"))

#contex locals
_payload_ctx_stack = LocalStack()

class MaxAPI(Namespace):
    @property
    def payload(self):
        '''Store the input payload in the current request context'''
        return _payload_ctx_stack.top

    @payload.setter
    def payload(self, value):
        if isinstance(value, dict):
            value = WrapperDict(value)

        _payload_ctx_stack.pop()
        _payload_ctx_stack.push(value)


MAX_API = MaxAPI('model', description='Model information and inference operations')


class Context(object):
    def __init__(self):
        self.app_name = None


context = Context()


class MAXApp(object):
    def __init__(self, title=API_TITLE, desc=API_DESC, version=API_VERSION):
        self.logger = getLogger()
        context.app_name = title
        self.app = Flask(title, static_url_path='')
        self.app.logger = getLogger()
        self.metrics = {}
        # load config
        if os.path.exists("config.py"):
            self.app.config.from_object("config")

        self.api = Api(
            self.app,
            title=title,
            description=desc,
            version=version,
            validate=True
        )

        self.api.namespaces.clear()
        self.api.add_namespace(MAX_API)

        @self.app.route("/metrics", methods=("GET",))
        def metrics():
            """
            监控接口
            :return:
            """
            resp = Response()
            resp.headers.set('Content-Type', 'text/plain')
            content = []
            for code in self.metrics.keys():
                pathData = self.metrics[code]
                for path in pathData.keys():
                    metricsData = pathData[path]
                    content.append("{}{{{}}} {}".format(
                        "count",
                        'path="{}",code="{}"'.format(path, code),
                        metricsData["count"]
                    ))
                    content.append("{}{{{}}} {}".format(
                        "total",
                        'path="{}",code="{}"'.format(path, code),
                        metricsData["total"]
                    ))
            resp.set_data("\n".join(content))
            return resp

        self.app.before_request(self.before)
        self.app.after_request(self.after)
        # enable cors if flag is set
        # if os.getenv('CORS_ENABLE') == 'true' and (
        #         os.environ.get('WERKZEUG_RUN_MAIN') == 'true' or self.app.debug is not True):
        # 默认跨域
        CORS(self.app, origins='*')
        print('NOTE: MAX Model Server is currently allowing cross-origin requests - (CORS ENABLED)')

    def add_api(self, api, route):
        MAX_API.add_resource(api, route)

    def mount_static(self, route):
        @self.app.route(route)
        def index():
            return self.app.send_static_file('index.html')

    def after(self, resp):
        ###参数错误情况下直接返回
        if resp.status_code == 400:
            print(resp.data)
            result = {'component_status': {'code': -1, 'message': ''}}
            data = eval(str(resp.data, 'utf-8'))
            message = ""
            for k, v in data["errors"].items():
                if v.endswith("is a required property"):
                    message = message + "missing parameter '" + k + "'"
                else:
                    n = v.find(" is not of type ")
                    if n >= 0:
                        message = message + "'" + k + "'" + v[n:]
                message += "; "
            message = message[:-2]
            result['component_status']['message'] = message

            resp.data = bytes(json.dumps(result), encoding='utf-8')
            return resp
        ###

        # resp.headers["trace-id"] = getTraceId()
        if request.path == "/metrics":
            return resp
        accessTime = request.__getattr__("access_time")
        done = int(time.time() * 1000) - accessTime
        statusCode = resp.status_code
        if statusCode not in self.metrics:
            self.metrics[statusCode] = {}
        if request.path not in self.metrics[statusCode]:
            self.metrics[statusCode][request.path] = {
                "count": 0,
                "total": 0
            }
        self.metrics[statusCode][request.path]["count"] += 1
        self.metrics[statusCode][request.path]["total"] += done
        return resp

    def before(self):
        request.__setattr__('access_time', int(time.time() * 1000))
        MAX_API.payload = request.get_json()
        '''
        batch mode，支持批处理，
        1. 判断header是否有X-batch-execute
        2. for 获取req body， 调用后端API
            a. 设置MAX_API.payload
            b. 调用flask dispatch_request分发请求
            c. 获取结果
        3. 合并结果
        4. 组装response，返回
        '''
        if request.headers.has_key("X-batch-execute"):
            self.logger.info("got header X-batch-execute, enter batch mode")
            batch_req = request.get_json()
            self.logger.info("batch request is : %s" % (batch_req))
            if not isinstance(batch_req, list):
                result = {
                    "stauts": "error",
                    "result": "wrong request body in batch mode. request body must be type of list"
                }
                response = self.app.make_response(result)
                return response

            batch_rv = []
            for req in batch_req:
                self.logger.debug("single request data is : %s" % (req))
                MAX_API.payload = req
                rv = self.app.dispatch_request()
                # 从response结果中获取data，bytes
                rv = rv.data
                rv_json = json.loads(bytes.decode(rv))
                self.logger.debug("single result data is : %s" % (rv_json))
                batch_rv.append(rv_json)
            # 根据data生成response
            self.logger.info("batch result is : %s" % (batch_rv))
            # 转换为json格式
            batch_rv = json.dumps(batch_rv)
            batch_rv = self.app.make_response(batch_rv)
            return batch_rv
        else:
            pass

    def run(self, host='0.0.0.0', port=5000):
        """
        启动服务
        :param host: host, 默认为0.0.0.0
        :param port: 端口, 默认为5000
        :return:
        """
        FlaskInstrumentor().instrument_app(self.app)
        RequestsInstrumentor().instrument()
        finalPort = self.app.config["PORT"] if "PORT" in self.app.config else port
        self.app.run(host=host, port=finalPort)

    def __call__(self, environ, start_response):
        return self.app(environ, start_response)



import datetime
import requests
import time
from cyclonefw.core.utils import getLogger

storageLogger = getLogger("storage")


def write(filename, data, timeout=600, service="http://oss.storage.svc.cluster.local"):
    """
    存储文件到文件存储服务器中
    :param timeout: 超时时间, 单位秒
    :param filename: 文件名称
    :param data: 文件数据,所有格式数据按照二进制处理
    :param service: 文件服务器地址, 在k8s中可不填写
    :return:
    """
    start = int(time.time())
    finalName = "{}_{}".format(filename, datetime.datetime.now().strftime("%Y-%m-%d_%H_%M_%S"))
    try:
        requests.post("{}/postUpload".format(service), headers={"filename": finalName}, data=data, timeout=timeout)
        storageLogger.info("文件: %s, id: %s, 大小: %s, 存储成功, 用时: %s", filename, finalName, len(data), int(time.time()) - start)
        return finalName
    except Exception as e:
        storageLogger.error("文件: %s 存储错误: %s", filename, e)
        raise e


def read(filename, timeout=600, service="http://oss.storage.svc.cluster.local"):
    """
    从文件存储中读文件
    :param timeout: 超时时间, 单位秒
    :param filename: 文件名称
    :param service: 文件服务器地址, 在k8s中可不填写
    :return:
    """
    start = int(time.time())
    try:
        resp = requests.get("{}/getFile?name={}".format(service, filename), timeout=timeout)
        storageLogger.info("文件: %s, 大小: %s, 读取成功, 用时: %s", filename, len(resp.content), int(time.time()) - start)
        return resp.content
    except Exception as e:
        storageLogger.error("文件: %s 读取错误: %s", filename, e)
        raise e
